(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived_multiple.
Require Import Derived_env.
Require Import Derived_arrsT.


(*************************************************************************)
(* Let expressions *)
(*************************************************************************)

(* let x : t = e1 in e2   :=   (\x : t. e2) e1 *)
Definition letE t e1 e2 : exp :=
  appE (lamE t e2) e1.

(* let x1 : t1 = e1 ... xN : tN eN in e   :=   (\x1 : t1 ... xN : tN. e) e1 ... eN *)
Definition letsE ts es e : exp :=
  appsE (lamsE ts e) es.


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Lemma substAE_let : forall a t t1 e1 e2,
  substAE a t (letE t1 e1 e2) = letE (substAT a t t1) (substAE a t e1) (substAE a t e2).
Proof.
  intros a t t1 e1 e2. unfold letE. simpl substAE. reflexivity.
Qed.

Corollary openAE_let : forall t t1 e1 e2,
  openAE t (letE t1 e1 e2) = letE (openAT t t1) (openAE t e1) (openAE t e2).
Proof.
  intros t t1 e1 e2. unfold openAE, openAT. apply substAE_let.
Qed.

Lemma substXE_let : forall x e' t1 e1 e2,
  substXE x e' (letE t1 e1 e2) = letE t1 (substXE x e' e1) (substXE (shift x) e' e2).
Proof.
  intros x e' t1 e1 e2. unfold letE. simpl substXE. reflexivity.
Qed.

Corollary openXE_let : forall e' t1 e1 e2,
  openXE e' (letE t1 e1 e2) = letE t1 (openXE e' e1) (substXE 1 e' e2).
Proof.
  intros e' t1 e1 e2. unfold openXE. replace (bnd 1) with (shift 0). apply substXE_let. reflexivity.
Qed.


Lemma substAE_lets : forall a t ts es e,
  substAE a t (letsE ts es e) = letsE (map (substAT a t) ts) (map (substAE a t) es) (substAE a t e).
Proof.
  intros a t ts es e. unfold letsE. 
  rewrite substAE_apps. rewrite substAE_lams. reflexivity.
Qed.

Corollary openAE_lets : forall t ts es e,
  openAE t (letsE ts es e) = letsE (map (openAT t) ts) (map (openAE t) es) (openAE t e).
Proof.
  intros t ts es e. unfold openAE, openAT. rewrite substAE_lets. reflexivity.
Qed.

Lemma substXE_lets : forall x e' ts es e,
  substXE x e' (letsE ts es e) = letsE ts (map (substXE x e') es) (substXE (shifts (length ts) x) e' e).
Proof.
  intros x e' ts es e. unfold letsE. rewrite substXE_apps. rewrite substXE_lams. reflexivity.
Qed.

Corollary openXE_lets : forall e' ts es e,
  openXE e' (letsE ts es e) = letsE ts (map (openXE e') es) (substXE (length ts) e' e).
Proof.
  intros e' ts es e. unfold openXE.
  rewrite substXE_lets. rewrite shifts_bnd. rewrite plus_zero. reflexivity.
Qed.


(*************************************************************************)
(* Multiple substitution *)
(*************************************************************************)

Lemma substsAE_let : forall as' ts t1 e1 e2,
  substsAE as' ts (letE t1 e1 e2) = letE (substsAT as' ts t1) (substsAE as' ts e1) (substsAE as' ts e2).
Proof.
  intros as' ts t1 e1 e2. unfold letE. rewrite substsAE_app. rewrite substsAE_lam. reflexivity.
Qed.

Corollary opensAE_let : forall ts t1 e1 e2,
  opensAE ts (letE t1 e1 e2) = letE (opensAT ts t1) (opensAE ts e1) (opensAE ts e2).
Proof.
  intros ts t1 e1 e2. do 3 rewrite unfold_opensAE. rewrite unfold_opensAT. apply substsAE_let.
Qed.

Lemma substsAE_lets : forall as' ts' ts es e,
  substsAE as' ts' (letsE ts es e) = letsE (map (substsAT as' ts') ts) (map (substsAE as' ts') es) (substsAE as' ts' e).
Proof.
  intros as' ts' ts es e. unfold letsE. 
  rewrite substsAE_apps. rewrite substsAE_lams. reflexivity.
Qed.

Corollary opensAE_lets : forall ts' ts es e,
  opensAE ts' (letsE ts es e) = letsE (map (opensAT ts') ts) (map (opensAE ts') es) (opensAE ts' e).
Proof.
  intros ts' ts es e. do 2 rewrite unfold_opensAE. rewrite substsAE_lets. f_equal.
  induction ts; simpl map.
    reflexivity.
    rewrite unfold_opensAT. rewrite IHts. reflexivity.
  induction es; simpl map.
    reflexivity.
    rewrite unfold_opensAE. rewrite IHes. reflexivity.
Qed.


(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Lemma lclosedE_let : forall L t1 e1 e2,
  lclosedT t1 ->
  lclosedE e1 ->
  (forall x, x \notin L -> lclosedE (openXE x e2)) ->
  lclosedE (letE t1 e1 e2).
Proof.
  intros L t1 e1 e2 Ht1 He1 He2.
  unfold letE.
  apply lclosedE_app.
    apply lclosedE_lam with (L := L).
      apply Ht1.
      intros x Fr. apply He2. apply Fr.
    apply He1.
Qed.

Lemma lclosedE_lets : forall L ts es e,
  lclosedTs ts ->
  lclosedEs es ->
  (forall xs, fresh L (length ts) xs -> lclosedE (opensXE (freeEs xs) e)) ->
  lclosedE (letsE ts es e).
Proof.
  intros L ts es e Hts Hes He.
  unfold letsE.
  apply lclosedE_apps.
    apply lclosedE_lams with (L := L).
      apply Hts.
      intros xs Frs. apply He. apply Frs.
    apply Hes.
Qed.


(*************************************************************************)
(* Typing *)
(*************************************************************************)

Lemma ofE_let : forall L D G t1 e1 t2 e2,
  ofE D G e1 t1 ->
  (forall x, x \notin L -> ofE D (G & x~t1) (openXE x e2) t2) ->
  ofE D G (letE t1 e1 e2) t2.
Proof.
  intros L D G t1 e1 t2 e2 He1 He2. unfold letE.
  eapply ofE_app.
    eapply ofE_lam.
      eapply validityTE. apply He1.
      apply He2.
    apply He1.
Qed.

Lemma ofE_lets : forall L D G ts es e t,
  length ts = length es ->
  (forall e' t', In (e', t') (combine es ts) -> ofE D G e' t') ->
  (forall xs, fresh L (length ts) xs -> ofE D (G & xs~~ts) (opensXE (freeEs xs) e) t) ->
  ofE D G (letsE ts es e) t.
Proof.
  intros L D G ts es e t Len Hes He. unfold letsE.
  eapply ofE_apps.
    eapply ofE_lams.
      intros t0 Elem.
        assert (Len' := Len). symmetry in Len'. destruct (in_combine_r_inv _ _ es ts t0 Len' Elem). 
        eapply validityTE. apply Hes. apply H.
      intros xs Frs. apply He. apply Frs.
    symmetry. apply Len.
    intros e' t' Elem. apply Hes. apply Elem.
Qed.
