(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import Metatheory_Misc.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived.
Require Import Modules_syntax.
Require Import Modules_semantic.
Require Import Modules_elaboration.
Require Import Modules_matching.


(*************************************************************************)
(* Let *)
(*************************************************************************)

Definition X : Id := 0.

(* let B in T  :=  {B; type X=T}.X *)
Definition LetT B T :=
  PathT (ModP (DotM (StrM (SeqB B (TypB X T))) X)).

(* let B in E  :=  {B; value X=E}.X *)
Definition LetE B E :=
  PathE (ModP (DotM (StrM (SeqB B (ValB X E))) X)).

(* let B in S  :=  {B; signature X=S}.X *)
Definition LetS B S :=
  PathS (ModP (DotM (StrM (SeqB B (SigB X S))) X)).

(* let B in M  :=  {B; module X=M}.X *)
Definition LetM B M :=
  DotM (StrM (SeqB B (ModB X M))) X.


Lemma elabT_let : forall L B T C G ks r e k t,
  C; G |=B B ~: exQ ks (strS r) ~> e ->
  (forall as' xs, fresh L (length ks) as' -> fresh L (sizeR r) xs ->
   (C & as'~~ks); (G &' (opensAR (freeTs as') r)@@xs) |=T T ~: k ~> t) ->
  ofT C t k ->
  C; G |=T LetT B T ~: k ~> t.
Proof.
Admitted.

Lemma elabE_let : forall L B E C G ks r e t e',
  C; G |=B B ~: exQ ks (strS r) ~> e ->
  (forall as' xs, fresh L (length ks) as' -> fresh L (sizeR r) xs ->
   (C & as'~~ks); (G &' (opensAR (freeTs as') r)@@xs) |=E E ~: t ~> e') ->
  ofT C t omK ->
  exists e'',
  C; G |=E LetE B E ~: t ~> e''.
Proof.
Admitted.


Lemma elabs_let : forall L B S C G ks r e q,
  C; G |=B B ~: exQ ks (strS r) ~> e ->
  (forall as' xs, fresh L (length ks) as' -> fresh L (sizeR r) xs ->
   (C & as'~~ks); (G &' (opensAR (freeTs as') r)@@xs) |=S S ~> q) ->
  ofT C (transQ q) omK ->
  C; G |=S LetS B S ~> q.
Proof.
Admitted.

Lemma elabM_let : forall L B M C G ks1 r ks2 s s' e e',
  C; G |=B B ~: exQ ks1 (strS r) ~> e ->
  (forall as1 xs, fresh L (length ks1) as1 -> fresh L (sizeR r) xs ->
   (C & as1~~ks1); (G &' (opensAR (freeTs as1) r)@@xs) |=M M ~: exQ ks2 s ~> e') ->
  (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
     length as2 = length ks2 ->
     opensAS (freeTs (as1 ++ as2)) s' = opensAS (freeTs as2) s) ->
  exists e'',
  C; G |=M LetM B M ~: exQ (ks1++ks2) s' ~> e''.
Proof.
Admitted.


(*************************************************************************)
(* Nested modules *)
(*************************************************************************)

Definition X1 : Id := 1.
Definition X2 : Id := 2.

(* M1 M2  :=  let module X1=M1; module X2=M2 in X1 X2 *)
Definition AppM' M1 M2 :=
  LetM (SeqB (ModB X1 M1) (ModB X2 M2)) (AppM X1 X2).

(* M:>S  :=  let module X=M in X:>S *)
Definition SealM' M S :=
  LetM (ModB X M) (SealM X S).

(* M:S  :=  (fun X:S -> X) M *)
Definition AnnM M S :=
  AppM' (FunM X S (IdM X)) M.

(* pack M:S  :=  let module X=M in pack X:S *)
Definition PackE' M S :=
  LetE (ModB X M) (PackE X S).


Lemma elabM_app' : forall L M1 M2 C G ks1 ks2 ks ts s1 s1' ks' s3 s3' s2 s2' e1 e2 f,
  C; G |=M M1 ~: exQ ks1 (funS ks s1 (exQ ks' s3)) ~> e1 ->
  C; G |=M M2 ~: exQ ks2 s2 ~> e2 ->
  (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
     length as1 = length ks1 ->
     opensAS (freeTs (as1 ++ as2)) (funS ks s1' (exQ ks' s3')) = opensAS (freeTs as1) (funS ks s1 (exQ ks' s3))) ->
  (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
     length as2 = length ks2 ->
     opensAS (freeTs (as1 ++ as2)) s2' = opensAS (freeTs as2) s2) ->
  (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
    matchS (C & as1~~ks1 & as2~~ks2) s2' s1' ks ts f) ->
  exists e'',
  C; G |=M AppM' M1 M2 ~: exQ (ks1++ks2++ks') (substsAS (map (shifts (length ks')) (bnds (length ks))) ts s3') ~> e''.
Proof.
Admitted.


Lemma elabM_seal' : forall L M S C G ks1 ks2 ts s1 s2 s2' e f,
  C; G |=M M ~: exQ ks1 s1 ~> e ->
  C; G |=S S ~> exQ ks2 s2 ->
  (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
     length as2 = length ks2 ->
     opensAS (freeTs (as1 ++ as2)) s2' = opensAS (freeTs as2) s2) ->
  (forall as1, fresh L (length ks1) as1 ->
    matchS (C & as1~~ks1) s1 s2' ks2 ts f) ->
  exists e'',
  C; G |=M SealM' M S ~: exQ ks2 s2 ~> e''.
Proof.
Admitted.

Lemma elabM_ann : forall L M S C G ks1 ks2 ts s1 s2 s2' s2'' e f,
  C; G |=M M ~: exQ ks1 s1 ~> e ->
  C; G |=S S ~> exQ ks2 s2 ->
  (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
     length as2 = length ks2 ->
     opensAS (freeTs (as1 ++ as2)) s2' = opensAS (freeTs as2) s2) ->
  (forall as1, fresh L (length ks1) as1 ->
    matchS (C & as1~~ks1) s1 s2' ks2 ts f) ->
  (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
     length as2 = length ks2 ->
     opensAS (freeTs as1 ++ ts) s2' = opensAS (freeTs as1) s2'') ->
  exists e'',
  C; G |=M AnnM M S ~: exQ ks1 s2'' ~> e''.
Proof.
Admitted.


(*************************************************************************)
(* Local *)
(*************************************************************************)

(* local B in D  :=  include (let B in {D}) *)
Definition LocalD B D :=
  InclD (LetS B (StrS D)).

(* local B1 in B2  :=  include (let B1 in {B2}) *)
Definition LocalB B1 B2 :=
  InclB (LetM B1 (StrM B2)).


Lemma elabD_local : forall L B D C G ks r e q,
  C; G |=B B ~: exQ ks (strS r) ~> e ->
  (forall as' xs, fresh L (length ks) as' -> fresh L (sizeR r) xs ->
   (C & as'~~ks); (G &' (opensAR (freeTs as') r)@@xs) |=D D ~> q) ->
  ofT C (transQ q) omK ->
  C; G |=D LocalD B D ~> q.
Proof.
Admitted.

Lemma elabB_local : forall L B1 B2 C G ks1 ks2 r s s' e1 e2,
  C; G |=B B1 ~: exQ ks1 (strS r) ~> e1 ->
  (forall as1 xs, fresh L (length ks1) as1 -> fresh L (sizeR r) xs ->
   (C & as1~~ks1); (G &' (opensAR (freeTs as1) r)@@xs) |=B B2 ~: exQ ks2 s ~> e2) ->
  (forall as1 as2, fresh L (length ks1 + length ks2) (as1 ++ as2) ->
     length as2 = length ks2 ->
     opensAS (freeTs (as1 ++ as2)) s' = opensAS (freeTs as2) s) ->
  exists e'',
  C; G |=B LocalB B1 B2 ~: exQ (ks1++ks2) s' ~> e''.
Proof.
Admitted.
